import os, importlib, logging
import torch
from lightning.pytorch.loggers import WandbLogger

from mura.repo.git_utils import understand_env
from mura.deploy.util import cprint
from mura.deploy.set_gpu import set_gpu

def run(args):
    
    version, action_id, task_id, run_id, paramfile, logfile = args[1:]
    
    logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
    py_logger = logging.getLogger('auto')
    py_logger.setLevel(logging.INFO)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    py_logger.addHandler(consoleHandler)

    fileHandler = logging.FileHandler(logfile)
    fileHandler.setFormatter(logFormatter)
    py_logger.addHandler(fileHandler)
        
    # load class 'param' from inside from paramfile.
    module_name = 'param'
    module_path = os.path.join(os.getcwd(), paramfile)
    
    if os.path.exists(module_path):
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        # Now you can use the module
    else:
        py_logger.warn(f"File not found: {module_path}")
    _config = module
    
    if hasattr(_config, 'config'):
        config = _config.config
        param = config.actions[action_id].tasks[task_id].run[run_id]
        action = config.actions[action_id]
        task = action.tasks[task_id]
        action_name = action.action_name
        task_name = task.task_name
        run_name = param.name if hasattr(param, 'name') else ''
        ngpu = action.ngpu
        project_name = config.project_name
    else:
        print(_config)
        param = _config.param
        action_name = ''
        task_name = ''
        run_name = param.name if hasattr(param, 'name') else ''
        ngpu = 1
        project_name = param.project_name
        
    param.gpus = set_gpu(ngpu)
    
    version, task, save_path = understand_env()
    fname = version + '.' + '.'.join([action_id, task_id, run_id])
    if action_name:
        fname += '-' + action_name
    if task_name:
        fname += '-' + task_name
    if run_name:
        fname += '-' + run_name

    # os.makedirs(os.path.join(save_path,"wandb/"), exist_ok=True)
    wandb_logger = WandbLogger(project=project_name, name=fname,
                    version=fname, config=param)
                    # save_dir=save_path) # has issues with sync.
                    
                    
    # TODO add watch stuff from train.py
                    
    return param, wandb_logger, py_logger, version, save_path


import subprocess, os

def parse_mem(x):
    return int(x.split(", ")[1].split(" MiB")[0])

def parse_gpu(x):
    return int(x.split(", ")[2].split(" %")[0])

cap = 1000 # MiB
ucap = 3 # %

def gpu_memory_usage():
    command = ["nvidia-smi", "--query-gpu=index,memory.used,utilization.gpu", "--format=csv"]
    output = subprocess.check_output(command).decode("utf-8").split("\n")[1:-1]
    print(f"Output: {output}")
    available = [(parse_mem(x) < cap) and (parse_gpu(x) < ucap) for x in output]
    indices = [int(x.split(", ")[0]) for x in output]   
    print(f"Available GPUs: {available}")
    # reindex to make sure that the indices are in order
    aid = [x for x in indices if available[indices.index(x)]]
    
    return aid

def set_gpu(ngpu):

    aid = gpu_memory_usage()
    os.environ['CUDA_VISIBLE_DEVICES'] = f"{','.join([str(x) for x in aid[:ngpu]])}"
    print(f"Using GPUs: {os.environ['CUDA_VISIBLE_DEVICES']}")
    return aid[:ngpu]

# # get a gpu in 0-4 and a gpu in 5-9
# def set_gpu(ngpu):

#     aid = gpu_memory_usage()
#     g1 = [x for x in aid if x < 5]
#     g2 = [x for x in aid if x >= 5]
#     print(g1, g2)
#     if ngpu == 1:
#         g1.extend(g2)
#         assert len(g1) > 0, "No GPU available"
#         os.environ['CUDA_VISIBLE_DEVICES'] = f"{g1[0]}"
#     else:
#         assert len(g1) > 0 and len(g2) > 0 , "No paired across system GPUs available"
#         os.environ['CUDA_VISIBLE_DEVICES'] = f"{g1[0]},{g2[0]}"



import os
import ffmpeg
import numpy as np
from tqdm import tqdm
import matplotlib.colors as mcolors
from einops import rearrange

import jpcm
from mura.draw.color import to_div

ndiv = np.array([jpcm.maps.rurikon,jpcm.maps.aoi,jpcm.maps.sora_iro,jpcm.maps.rgb(255,255,255)]) # pretty good right now
cmap = to_div(ndiv,rot=1/3,net_rot=0.0,sat_factor=100)
# cmap = to_div(ndiv,rot=1/3+0.025/3,net_rot=-0.025,sat_factor=100)

def fast_scandir(dirname):
    subfolders= [f.path for f in os.scandir(dirname) if f.is_dir()]
    for dirname in list(subfolders):
        subfolders.extend(fast_scandir(dirname))
    return subfolders

def mp4(filename, d, fps=2, triplet=True, mn = None, cmap=cmap):
    """Saves a NumPy array of frames as an MP4 video."""
 
    if d.shape[0] > 4000:
        d = d[1000:3000,...]
 
 
    if triplet:
        frames = np.stack([ d[:,0,...], d[:,0,...] + d[:,1,...],d[:,1,...]], axis=1).astype(np.float32) # x_err, xhat, x
        print(frames.shape)
        r = np.maximum(np.max(frames[:,-1], axis=(0,2,3)),-np.min(frames[:,-1],axis=(0,2,3)))[None,None,:,None,None] # B, Y, C, H, W
        nframes = frames / (2*r) + 0.5
        nframes = rearrange(nframes,'b y c h w -> b (c h) (y w)')
        print(np.max(nframes), np.min(nframes))
    else:
        assert mn is not None
        assert d.shape[1] == mn[0] * mn[1]
        frames = d
        r = np.maximum(np.max(frames, axis=(0,2,3)),-np.min(frames,axis=(0,2,3)))[None,:,None,None] # B, C, H, W
        nframes = frames / (2*r) + 0.5
        nframes = rearrange(nframes,'b (c d) h w -> b c d h w', c=mn[0], d=mn[1])
        nframes = rearrange(nframes,'b c d h w -> b (d h) (c w)')
        print(np.max(nframes), np.min(nframes))
        
    print(cmap)
    frames = cmap(nframes)[...,:3]
    frames *= 255 / np.max(frames)
    n, height, width, channels = frames.shape
    print(frames.shape)
    process = (
        ffmpeg
        .input('pipe:', format='rawvideo', pix_fmt='rgb24', s=f'{width}x{height}')
        .output(filename, pix_fmt='yuv420p', vcodec='libx264', r=fps)
        .overwrite_output()
        .run_async(pipe_stdin=True)
    )
    for frame in frames:
        process.stdin.write(frame.astype(np.uint8).tobytes())
    process.stdin.close()
    process.wait()


if __name__ == '__main__':    

    names = ['valid','infer']
    directory = 'run/'
    remake = True #False

    print('scan directory')
    gen = fast_scandir(directory)
    for path in tqdm(gen):
        fname = os.path.join(path,'valid.mp4')
        qname = os.path.join(path,'infer.npz')
        if (not os.path.exists(fname) and os.path.exists(qname)) \
              or (os.path.exists(qname) and remake):
            print(path)
            try:
                data = [np.load(os.path.join(path,f"{x}.npz"))['arr_0'] for x in names]    
                for name, d in zip(names, data):
                    fname = os.path.join(path,f'{name}.mp4')
                    mp4(fname, d)
            except Exception as e:
                print(e)

# run this script with 
# conda activate ./sw
# source .venv/bin/activate
# python3 src/ml/runners/basic_plot.pyWW


import os
import git
import yaml
import logging
import time
logger = logging.getLogger('GIT')

## TODO need to figure out auto-push when SSH keys are not active... (both normally and in container)

class git_utils:
    
    def __init__(self, **kwargs):
        self.__name__ = 'git_utils'

        self.git_enabled = not os.environ.get('IS_CONTAINER', False)
        self.whoami = os.environ.get('USER', 'guest')

        self.local_dev = f'auto-dev-{self.whoami}'
        self.local_run = f'auto-run-{self.whoami}'
        self.version_file = kwargs.get('version_file','auto/version.yml')
        # get auto folder from version file
        self.auto_folder = os.path.dirname(self.version_file)
        os.makedirs(self.auto_folder, exist_ok=True)
        
        self.repo_path = kwargs.get('repo_path','.')

        self.skip_checks = kwargs.get('skip_checks', ['param.py'])
        self.skip_checks.extend([self.version_file])
        self._checks = kwargs.get('checks', ['src/ml/models/', 'src/ml/models/layers/', 'src/ml/models/connections/', 'src/ml/models/blocks/', 'src/ml/models/act/', 'src/ml/param/', 'src/ml/runners/', 'src/ml/iox/'])

    def repo(self,strict=True):
        os.system('wandb login')
        try:
            repo = git.Repo(self.repo_path)
            self.ensure_local_dev_branch(repo)
            if strict:
                self.on_dev(repo)
                if repo.is_dirty():
                    logger.warning(f"Repository is dirty. Committing changes.")
                    repo.git.add(all=True)
                    msg = input("MSG: ")
                    repo.git.commit(message=f'[auto] {msg}')
            return repo
        except (git.exc.InvalidGitRepositoryError):
            repo = git.Repo.init(None)
            repo.config_writer().set_value('push', 'followTags', 'true').release()
            self.checkout_local_dev_branch(repo)
            return repo

    # Make sure you have a repository, and there is a remote named {remote_name} and a branch named {branch_name}.
    # The easiest way to fix this is to clone the repository (from the appropriate tagged release).""")

    def on_dev(self, repo):
        assert self.local_dev == repo.head.ref.name, f"Expected to be on {self.local_dev}, but on {repo.head.ref.name}. Switch first."
            
    def checkout_local_dev_branch(self, repo):
        self.ensure_local_dev_branch(repo)
        repo.git.checkout(self.local_dev) 

    def ensure_local_dev_branch(self, repo):
        if self.local_dev not in repo.branches:
            repo.create_head(self.local_dev)
        
    def source_modified(self, repo, expected_commit_hash):
        # check if local_dev is ahead of local_run
        if self.local_run not in repo.branches:
            return True
        commit1 = repo.heads[self.local_dev].commit
        commit2 = repo.heads[self.local_run].commit
        
        # check if commit hash has changed
        commit_hash = repo.heads[self.local_dev].commit.hexsha
        commit_hash_changed = commit_hash != expected_commit_hash
        
        # check diffs
        diff = repo.git.diff(name_only=True).split('\n') # get diff against index
        diff = [d for d in diff if any([s in d for s in self._checks])]
        logger.info(f"Source diff: {diff}")

        # create state modified flags
        smb = (len(diff) > 0) or commit_hash_changed # source modified
        sm = commit1 != commit2 or smb              # source modified strict (incl. run commit)
        
        return sm, smb, commit_hash
        
    def load_version(self, f):
        if not os.path.exists(f):
            version_data = {
                'version': [0,0,0],
                'commit': '0'*40,
                'action': 0,
                'total_actions': 0
            }
            self.save_version(f, version_data)
        version_data = yaml.load(open(f).read(), Loader=yaml.FullLoader)
        return version_data

    def save_version(self, f, version_data):
        with open(f,'w') as file:
            yaml.dump(version_data, file)
        
    def remove_index_lock(self):
        if os.path.exists('.git/index.lock'):
            logger.warn('Automatically removing .git/index.lock (in 2s); please make sure no other git process is running.')
            time.sleep(2)
            os.remove('.git/index.lock')    
            
    def tag_and_version(self, repo, tag=True):
        if self.local_run not in repo.branches:
            repo.create_head(self.local_run)
            
        version_data = self.load_version(self.version_file)    
        expected_commit_hash = version_data['commit']
        sm, smb, commit_hash = self.source_modified(repo, expected_commit_hash)
            
        if (sm and tag) or smb:
            version = version_data['version']
            version[2] += 1
            version_data['version'] = version
            version_data['commit'] = commit_hash
            version_data['action'] = 0
        # update task number
        else:
            version_data['action'] += 1
        version_data['total_actions'] += 1
        self.save_version(self.version_file, version_data)
            
        try:        
            repo.git.add('-u') # update index for next diff check
            self.remove_index_lock() # remove again just in case
        except Exception as e:
            self.remove_index_lock()
            repo.git.add('-u')
            self.remove_index_lock() # remove again just in case    
        
        if tag:
            try:
                if sm:
                    repo.git.add(self.version_file)
                    repo.git.commit(message=f'[auto] version update')
                    repo.git.checkout(self.local_run)
                    repo.git.merge(self.local_dev)
                    tag = '.'.join([str(v) for v in version])
                    repo.create_tag(tag)
                    repo.git.checkout(self.local_dev)
            except Exception as e:
                logger.error(f"Failed to update git and tag: {e}")
                repo.git.checkout(self.local_dev) 
        return version_data, sm

    def get_save_path(self, version_data, strict, action_name, base_save_path):
        vs = [str(v) for v in version_data['version']]
        version = f'v-{".".join(vs)}{"" if strict else "a"}'
        fname =  f'{version}-{version_data["action"]}-{action_name}'
        sp = os.path.join(base_save_path, fname)
        logger.info(f'Save path: {sp}')
        os.makedirs(sp, exist_ok=True)
        return version, str(version_data["action"]), sp

    def save_env(self, v, a, sp):
        env = {
            'version': v,
            'action': a,
            'save_path': sp
        }
        os.environ.update(env)   
        
def understand_env():
    return (
        os.environ.get('version', '0.0.0'),
        os.environ.get('action', 0),
        os.environ.get('save_path', 'save')
    )


import subprocess
import time

from mura.services.wandb_sync import sync_wandb
    
def loop(start_time, functions, min_time=200, t=2): # in seconds
    running = True
    while running or time.time() - start_time < min_time:
        for f in functions:
            f()
        running = len(subprocess.check_output("qstat")) > 0
        time.sleep(t)
        
if __name__ == "__main__":
    start_time = time.time()
    loop(start_time, [sync_wandb])


import logging, os, copy

from mura.deploy.util import cupdate, cselect, serialize_class, check_wifi
from mura.deploy.templates import sg_engine, Bash, sg_engine_filename

code_folder = '.src'
run_folder = 'run'
run_info_file = 'param.py'

service_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../services/service_tmux.py')

## logging
logFormatter = logging.Formatter("%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
rootLogger = logging.getLogger('auto-deploy-action')
rootLogger.setLevel(logging.INFO)

consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)
rootLogger.addHandler(consoleHandler)   

tasklogger = rootLogger.getChild('task')


def deploy(deploy_function, action_info, index, key, parent):
    for i, task in enumerate(action_info.__dict__[key]):
        task_info = copy.deepcopy(task)
        cupdate(task_info, action_info, [key])
        deploy_function(task_info, [*index,i], parent)

def deploy_action(action_info, i, parent, _gl):

    repo = _gl.repo(strict=action_info.strict_version_checks)
    version_data, source_modified_flag = _gl.tag_and_version(repo, tag=action_info.strict_version_checks)
    version, action, save_path = _gl.get_save_path(version_data, action_info.strict_version_checks, action_info.action_name, action_info.base_save_path)
    _gl.save_env(version, action, save_path)

    fileHandler = logging.FileHandler(f"{save_path}/version.log")
    fileHandler.setFormatter(logFormatter)
    rootLogger.addHandler(fileHandler)
    logger = rootLogger
        
    hostname = os.uname()[1]
    logger.info(f'hostname: {hostname}')
    code_path = os.path.join(save_path,code_folder)
    
    action_info.save_path = save_path
    action_info.code_path = code_path
    action_info.scripts_path = os.path.join(code_path, run_folder)
    action_info.source_modified_flag = source_modified_flag
    action_info.version_data = version_data
    action_info.action_number = action # not i! 
    action_info.hostname = hostname

    os.makedirs(action_info.code_path, exist_ok=True)
    os.makedirs(action_info.scripts_path, exist_ok=True)
    
    # copy all code to task folder
    # assumes 'src' is named the same as the project name
    os.system(f'cp -r * {action_info.code_path}')   
    
    deploy(deploy_task, action_info, i, 'tasks', parent) # need to make these dependent on previous action



def get_param(i, parent):
    q = copy.deepcopy(parent)
    parts = []
    u = q
    for k, item in zip(i, ['actions', 'tasks', 'runs']):
        u = u.__dict__[item][k]
        parts.append(u)
        
    u = q
    for part,item in zip(parts,['actions', 'tasks', 'runs']):
        setattr(u, item, [part,])
        u = part
    
    return q

def deploy_task(task_info, i, parent):
    
    task_id = i[-1]
    task_info.__name__ = 'task'
    file_ = f"{task_id}"
    if task_info.task_name:
        file += f'-{task_info.task_name}'
    task_info.task_path = os.path.join(task_info.save_path, file_)
    os.makedirs(task_info.task_path, exist_ok=True)
    with open(os.path.join(task_info.task_path,'readme.md'), 'w') as f:
        f.write(task_info.__dict__.get('description', ''))
        
    param = get_param(i, parent)
    
    deploy(deploy_run, task_info, i, 'runs', param)           


def deploy_run(run_info, _id, system):
    # now start each run
   
    run_info.action_id, run_info.task_id, run_info.run_id = _id
    action = system.actions[_id[0]]
    run_info.__name__ = 'run'
    run_info.version = action.version_data['version']
    run_info._id = _id
    run_info.job_name = f'{system.project_name}-v{".".join([str(s) for s in run_info.version])}-{action.action_number}.{run_info.task_id}.{run_info.run_id}'

    run_info.save_path = os.path.join(run_info.task_path, str(run_info.run_id))
    os.makedirs(run_info.save_path, exist_ok=True)
    
    with open(run_info.save_path + '/' + run_info_file, 'w') as f:
        f.write(str(system))

    serialize_class(system, run_info.save_path + '/' + run_info_file)

    os.environ['save_path'] = run_info.save_path
    logfile = f'{run_info.save_path}/.log'
    
    
    with Bash(sg_engine(system.run, run_info,
                        pylogfile=logfile.replace('.log', '_python.log'),
                        logfile=logfile.replace('.log', '_engine.log')),
              path=run_info.save_path):
        
        wifi = check_wifi()
        
        if not wifi:
            os.environ["WANDB_MODE"] = "offline" # disable wandb sync  
        
        if action.no_compute or (action.hostname != system.cluster_name):
            tasklogger.info('Starting local task...')
            os.system(f'source {sg_engine_filename}')
        else:
            tasklogger.info('Starting cluster task...')
            os.system(f'qsub {sg_engine_filename}')
        
            if not wifi:
                tasklogger.info('Starting services...')
                # os.system(f"tmux kill-session -t {run_info.project_name}-services")
                # os.system(f"tmux new -A -d -s {run_info.project_name}-services 'python3 src/auto/services/service_tmux.py; $SHELL'")  
        
    #     tasklogger.info('Task submitted.')
            
### add to a queue and run actions serially 
### run tasks in parallel

# Problem with sg-queue is that GPUs are not exclusive etc...
# but you can cancel jobs though...

# so use sg-queue with smart GPU allocation on backend
# and then run tasks in parallel per action...with delayed task start to get correct gpu
# alternatively, use a queue system that has GPU exclusivity built in...


import os
# from schema import Schema, And, Use, Optional, Or

from mura.deploy.util import crupdate, cprint
from mura.deploy.deploy_action import deploy, deploy_action
from mura.repo.git_utils import git_utils as gl

# param_schema = Schema({
#     'grid': {
#         'Nx': int,
#         'Ny': int,
#         'Lx': float,
#         'Ly': float,
#     },
# })
                    
class run:
    env_init_template = 'sg_env_init.jinja'
    env_end_template = 'sg_env_end.jinja'
    submit_template = 'sg_engine.jinja'    ## jobname, logfile, run_commands
    run_template = 'run.jinja'          ## env_init_commands, env_end_commands, {{runner}} {{run}} {{configfile}} {{logfile}}
    param_file = 'param.py'
    
    install_info = 'install'

class action:
        action_name = 'action'
        no_compute = False
        strict_version_checks = False
        base_save_path = '../run/' # outside source folder
        #----
        
        debug = False
        save_frequency = 100 # only for ML
        ngpu = 1
        data = "" # only for ML; data folder

class DefaultScheduler():    
    def __init__(self, _params):
        self.__name__ = 'config'
        self.run = run
        self.actions = \
            [
                action,
            ]

        if _params is not None:
            crupdate(self,_params)

class Instancer():
    def __init__(self, run_validator, config=None, gconfig=None):
        self.__name__ = 'Instancer'
        self.defaults = DefaultScheduler(config)
        gdict = gconfig.__dict__ if gconfig is not None else {}
        self.gl = gl(**gdict, repo_path=os.getcwd())
        
        if run_validator is not None:
            self.validate(run_validator)
        
    def validate(self, run_validator):
        # go into each action, task, run, and validate
        for i,_action in enumerate(self.defaults.actions):
            for j,_task in enumerate(_action.tasks):
                for k,_run in enumerate(_task.runs):
                    print(_task, _run)
                    self.defaults.actions[i].tasks[j].runs[k] = run_validator(_run)
        
    def run(self):
        deploy(lambda *x: deploy_action(*x,self.gl), self.defaults, [], 'actions', self.defaults)

            
    def __repr__(self):
        return f'Scheduler: \n\t{cprint(self.defaults,level=1)}\nGit: \n\t{cprint(self.gl,level=1)}'
        
            
class SingleInstancer(Instancer):
    def __init__(self, run_validator, config, param=None, gconfig=None, _task_name='', _action_name='', **action_kwargs):
        class task:
            task_name = _task_name
            runs = [param,]
        class action:
            action_name = _action_name
            tasks = [task,]
    
        for key, value in action_kwargs.items():
            setattr(action, key, value)    
        
        config.actions = [action,]        
        super().__init__(run_validator=run_validator, config=config, gconfig=gconfig)


import hashlib
import importlib.util
import inspect
import re 
import os 
import time

def consistent(d1,d2):
    for k in d2.keys():
        if 'train' in k or 'test' in k:
            continue
        assert d1.__dict__[k] == d2[k], f'{k} mismatch: {d1.__dict__[k]} != {d2[k]}; retraining model from scratch'

def is_local_module(module_name):
    spec = importlib.util.find_spec(module_name)
    return spec is not None \
        and spec.origin is not None \
        and os.path.dirname(__file__) in spec.origin
        
def get_imports(module_file_header):
    # find any other files that are imported by the model file
    module_file = module_file_header
    if 'util' in module_file:
        return []
    
    with open(os.path.join(os.getcwd(),f'{module_file}.py'), 'r') as f:
        module_code = f.read()
    imports = re.findall(r'from (.*?) import', module_code)
    imports += re.findall(r'import (.*?)\n', module_code)
    # select imports that are not standard
    return [i.replace('.','/') for i in imports if is_local_module(i)]    

def get_all_imports(module_names, imports):
    for module_name in module_names:
        cim = get_imports(module_name)
        imports.extend(cim)
        if len(cim) > 0 :
            get_all_imports(cim, imports)
    return imports
    

def capture_local_module(module_name):
    # get all modules
    module_file = inspect.getfile(module_name).replace('.py','').replace(os.getcwd(),'')[1:]
    modules = get_all_imports([module_file], [])
    modules.extend([module_file])
    return modules

time_parse = lambda t: time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(t))        

def time_stat(model):
        
    cpath = os.getcwd()
    keys = capture_local_module(model)
    hashes = {
        k: hashlib.md5(open(os.path.join(cpath,f'{k}.py'),'rb').read()).hexdigest() for k in keys
    }
    ts = [os.path.getmtime(os.path.join(cpath,f'{k}.py')) for k in keys]
    times = {
        k: time_parse(t) for k,t in zip(keys,ts)
    }
    last_save_time = time_parse(max(ts))
    return [last_save_time, hashes, times]


import jpcm
import numpy as np
import matplotlib
import colour

# TODO add to jpcm library instead

def to_div(ndiv, rot=0, net_rot=0, sat_factor=1.0, strength=1.0, peak_whiten=1.0, cut=0):
  ### Linear Models
  # to REC 706 # need to confirm this is the right space...
  # yc_mat = np.array([[0.2126, 0.7152, 0.0722],
  #                   [-0.1146, -0.3854, 0.5],
  #                   [0.5, -0.4542, -0.0458]])
  # rgb_mat = np.array([[1, 0, 1.5748],
  #                     [1, -0.1873, -0.4681],
  #                     [1, 1.8556, 0]])

  # to REC 2020
  # kr = 0.2627
  # kg = 0.678
  # kb = 0.0593

  # # stupid colormap
  # kr = 1/3
  # kg = 1/3
  # kb = 1/3

  # yc_mat = np.array([[kr, kg, kb],
  #                   [-0.5 * (kr / (1-kb)), -0.5 * (kg / (1-kb)), 0.5],
  #                   [0.5, -0.5 * (kg / (1-kr)), -0.5 * (kb / (1-kr))]])
  
  # rgb_mat = np.array([[1, 0, 2 * (1-kr)],
  #                     [1, -2 * (kb/kg) * (1-kb), -2 * (kr/kg) * (1-kr)],
  #                     [1, 2 * (1-kb), 0]])

  # rgb_mat = np.linalg.inv(yc_mat)

  # nybr = ndiv @ yc_mat.T

  # diff = [nybr[i] - nybr[i-1] for i in range(len(nybr)-1,0,-1)]
  # pybr = np.cumsum([nybr[-1],*diff], axis=0) 
  # pybr[:,0] = np.flip(nybr[:,0])
  # print(nybr,'\n',pybr)

  # pdiv = np.fabs(pybr @ rgb_mat.T)

  ### Nonlinear models

  nhsl = colour.RGB_to_HSL(ndiv)
  # rotate white / black
  sat = nhsl[:,1]
  wh = np.where(sat < 1e-6)
  nhsl[wh,0] = rot % 1

  # adjust all
  nhsl[:,0] = (nhsl[:,0] + net_rot) % 1

  # stretch lighting
  mu = np.mean(nhsl[:,2])
  nhsl[:,2] = np.clip((nhsl[:,2] - mu) * strength  + mu, 0, 1)
  
  # oversaturate
  nhsl[:,1] = np.clip(nhsl[:,1] * sat_factor, 0, 1)

  # peak whiten
  nhsl[-1,2] = np.clip(nhsl[-1,2] * peak_whiten, 0, 1)

  phsl = np.flip(np.copy(nhsl), axis=0) # same lightness and sat.
  diff = [(nhsl[i,0] - nhsl[i-1,0]) for i in range(len(phsl)-1,0,-1)]
  phsl[:,0] = np.cumsum([nhsl[-1,0],*diff], axis=0)
  
  # phsl[:,0] = (0.5 + phsl[:,0]) % 1
#   print(nhsl,'\n-\n',phsl)
  end = len(ndiv)
  ndiv = colour.HSL_to_RGB(nhsl[:end-cut])
  pdiv = colour.HSL_to_RGB(phsl[cut:])

#   print(ndiv,'\n',pdiv)


  n_cmp = (jpcm.core.get('div0',cmaps={
      'div0':ndiv
  })[0])

  p_cmp = (jpcm.core.get('div0',cmaps={
    'div0':pdiv
  })[0])

  assert len(n_cmp) == len(p_cmp), f"{len(n_cmp)} != {len(p_cmp)}"

  cmap = matplotlib.colors.ListedColormap([*n_cmp,*p_cmp])
  return cmap



import copy

def update(d1,d2,skip):
    # update d1 with d2, skipping keys in skip
    d3 = {k:v for k,v in d2.items() if k not in skip}
    d1.update(d3)    

def cupdate(c1,c2,skip):
    # update c1 with c2, skipping keys in skip
    # c1, c2 are classes
    for key, value in c2.__dict__.items():
        if not key.startswith('__'):
            if key not in skip:
                setattr(c1, key, value)
    
def crupdate(_A, _params): 
    # classes though, so this is a shallow/deep/mixed copy
    def _update(A,B):
        # if isinstance(B, dict):
        #     for key, value in B.items():
        #         if not key.startswith('__'):
        #             if key in A.keys():
        #                 _update(A[key], value)
        #             else:
        #                 setattr(A, key, value)
                        
        if isinstance(B, list) and isinstance(A, list):
            for i in range(len(B)):
                if i < len(A):
                    if type(A[i]) not in [str, int, float, bool]:
                        _update(A[i], B[i])
                    else:  
                        A[i] = B[i]
                else:
                    if type(A[-1]) not in [str, int, float, bool]:
                        A[i] = copy.deepcopy(A[-1])
                        _update(A[i], B[i])
                    else:  
                        A[i] = B[i]
                    
        elif hasattr(B, '__dict__'):
            for key, value in B.__dict__.items():
                if not key.startswith('__'):
                    if key in A.__dict__.keys() and type(value) not in [str, int, float, bool]:
                        _update(A.__dict__[key], value)
                    else:
                        setattr(A, key, value)
        else:
            raise ValueError(f'Cannot update {A} with {B}')
                    
    _update(_A, _params)

def pretty(d, indent=0, out=[]):
    for key, value in d.items():
      out += ('\t' * indent + str(key))
      if isinstance(value, dict):
         pretty(value, indent+1)
      else:
         out += ('\t' * (indent+1) + str(value))
    return ''.join(out)
    
def internal_update(d, u):
    for k, v in u.items():
        if isinstance(d, list):
            for i in range(len(d)):
                d[i] = internal_update(d[i], u)
        elif k in d.keys():
            d[k] = v
        else:
            for kd in d.keys():
                d[kd] = internal_update(d[kd], u)
    return d
    
def select(d,i, k):
    d2 = copy.deepcopy(d)
    
    dc = d2
    for key in k[:-1]:
        dc = dc[key][0]
    dc = {k[-1]:[dc[k[-1]][i],]} # dc is now the dictionary we want to keep
    
    d2 = internal_update(d2, dc)
    return d2

def cselect(c,u,k):
    c2 = copy.deepcopy(c)
    c2.__dict__[k] = [u,]
    return c2

def cprint(c,level=0):
    if isinstance(c, dict):
        return '\n' + '\n'.join(['\t'*level + f'{key}={cprint(value, level+1)}' for key, value in c.items()])
    elif isinstance(c, list):
        return '\n' + '\n'.join(['\t'*level + cprint(value, level+1) for value in c])
    elif hasattr(c, '__dict__'):
        return '\n' + '\n'.join(['\t'*level + f'{key}:{cprint(value, level+1)}' for key, value in c.__dict__.items() if not key.startswith('__')])
    else:
        return str(c)
    
    
def serialize_class(cls, file_name):
    def process_class(cls, indent_level=0):
        _indent = "    "
        indent = _indent * indent_level
        _name = cls.__name__ if hasattr(cls, '__name__') else cls.__class__.__name__
        if _name.startswith('<'):
            return ''
        # class_code = f"{indent}class {_name}:\n"
        class_code = f"type('{_name}', (object,), {r'{'} \n"
        
        # Check for attributes in the class (via __annotations__ for typed attributes)
        annotations = False
        for key, value in cls.__dict__.items():
            if not key.startswith('__'):
                if hasattr(value, '__dict__'):  # Check if it's a class
                    class_code += f"{indent}'{value.__name__}':" + process_class(value, indent_level + 1)
                elif isinstance(value, list):
                    class_code += f"{indent}    '{key}':[ \\\n"
                    if hasattr(value[0], '__dict__'):
                        for i, item in enumerate(value):
                                class_code += _indent * (indent_level + 2) + process_class(item, indent_level + 2)
                                if i < len(value)-1:
                                    class_code += ", \n"
                    else:
                        for i, item in enumerate(value):
                            if isinstance(item, str):
                                item = f"'{item}'"
                            class_code += f"{indent}        {item},\n"
                    class_code += f"{indent}    ],\n"
                else:
                    if isinstance(value, str):
                        value = f"'{value}'"
                    class_code += f"{indent}    '{key}':{value},\n"

        # # Add a placeholder if the class is empty
        # if not annotations and not any(isinstance(obj, type) for obj in cls.__dict__.values()):
        #     class_code += f"{indent}    pass\n"
        
        class_code += f"{indent} {r'}'}),\n"
        
        return class_code

    # Generate code for the main class and its nested structure
    class_code = process_class(cls)

    # Write to the given .py file
    with open(file_name, 'w') as file:
        file.write(class_code)
        
        
import socket

def check_wifi(host="www.google.com", port=80, timeout=3):
    """
    Checks if a Wi-Fi connection is available by attempting to connect to a host.
    
    Args:
        host (str, optional): The hostname to connect to. Defaults to "www.google.com".
        port (int, optional): The port number to connect to. Defaults to 80.
        timeout (int, optional): Timeout in seconds for the connection attempt. Defaults to 3.
    
    Returns:
        bool: True if a connection is established, False otherwise.
    """
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error:
        return False


import os
from jinja2 import Environment, FileSystemLoader

from mura.util import localpath

env = Environment(loader=FileSystemLoader(localpath('templates/')),comment_start_string='###',line_statement_prefix = '###',line_comment_prefix='###')

script_folder = 'auto/tmp'
sg_engine_filename = 'auto/tmp/sg_engine.sh'
os.makedirs(script_folder, exist_ok=True)

def sg_engine(run, runinfo, **kwargs):
    template = env.get_template(run.submit_template)
    _init = env.get_template(run.env_init_template).render(
        install=run.install_info,
        **kwargs)
    _end = env.get_template(run.env_end_template).render(**kwargs)
    atr_id = runinfo._id
    single_run = env.get_template(run.run_template).render(
        env_init_commands=_init,
        env_end_commands=_end,
        runner = runinfo.runner,
        version = '.'.join([str(s) for s in runinfo.version]),
        a_index = atr_id[0],
        t_index = atr_id[1],
        r_index = atr_id[2],
        configfile = run.param_file,
        **kwargs)  
    return template.render(**kwargs,
                           run_commands=single_run,
                           jobname = runinfo.job_name,
                           )

def write_script(filename, script):
    os.makedirs(script_folder, exist_ok=True)
    with open(filename, 'w') as f:
        f.write(script)
    os.system(f'chmod +x {filename}')

class Bash():
    def __init__(self, script, path='.', delete=False, **kwargs):
        self.filename = sg_engine_filename
        self.script = script
        self.delete = delete
        self.path = path
        self.kwargs = kwargs
    def __enter__(self):
        write_script(self.filename, self.script)
        # cp script to path
        os.system(f'cp {self.filename} {self.path}')
        
    
    def __exit__(self, *args, **kwargs):
        if self.delete:
            sleep(4)
            self.delete_script()
    
    def delete_script(self):
        assert self.filename is not None
        assert os.path.exists(self.filename)
        assert self.filename.contains('/tmp/')
        os.system(f'rm {self.filename}')